import unittest
import os
import pyosmapi.osm_api
from rauth import OAuth1Service
from requests_oauthlib import OAuth1


class MyTestCase(unittest.TestCase):

    def test_oauth(self):
        try:
            CONSUMER_KEY = os.environ['OSM_CONSUMER_KEY']
            CONSUMER_SECRET = os.environ['OSM_CONSUMER_SECRET']
        except KeyError:
            self.skipTest("external resource not available")
            exit()

        osm_auth = OAuth1Service(
            name='pyosmapi',
            consumer_key=CONSUMER_KEY,
            consumer_secret=CONSUMER_SECRET,
            request_token_url='https://master.apis.dev.openstreetmap.org/oauth/request_token',
            access_token_url='https://master.apis.dev.openstreetmap.org/oauth/access_token',
            authorize_url='https://master.apis.dev.openstreetmap.org/oauth/authorize',
            base_url='https://master.apis.dev.openstreetmap.org')

        request_token, request_token_secret = osm_auth.get_request_token()
        authorize_url = osm_auth.get_authorize_url(request_token)
        print('Visit this URL in your browser: ' + authorize_url)
        input('press any key to continue...')  # `does not need
        acc_token, acc_token_secret = osm_auth.get_access_token(request_token, request_token_secret)

        print('token: ' + acc_token + ' secret: ' + acc_token_secret)
        req_token = OAuth1(osm_auth.consumer_key, osm_auth.consumer_secret, acc_token, acc_token_secret)
        api = pyosmapi.osm_api.OsmApi()
        user = api.get_current_user(req_token)
        print(user)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from pyosmapi import exceptions
import pyosmapi.osm_api as osmapi
import pyosmapi.osm_util


class MyTestCase(unittest.TestCase):
    osmo = osmapi.OsmApi()

    def auth(self):
        try:
            username = os.environ['OSM_USERNAME']
            password = os.environ['OSM_PASSWORD']
            return username, password
        except KeyError:
            exit()

    def test_get_api_versions(self):
        data = self.osmo.get_api_versions()
        print(data)

    def test_capability(self):
        data = self.osmo.get_api_capabilities()
        print(data)

    def test_permissions(self):
        data = self.osmo.get_permissions(self.auth())
        print(data)

    def test_cs_cre(self):
        data = self.osmo.open_changeset({'comment': 'test', 'created_by': 'osmate'}, self.auth())
        print(data)

    def test_cs_unsub(self):
        data = self.osmo.unsub_changeset(177967, self.auth())
        print(data)

    def test_cs_sub(self):
        data = self.osmo.sub_changeset(177967, self.auth())
        print(data)

    def test_cs_get(self):
        cs = self.osmo.get_changeset(177967)
        print(cs)

    def test_cs_get_by_params(self):
        cs = self.osmo.get_changesets(bbox=(13.428416654163087, 52.49863874116848, 13.446383345836914, 52.52816125883152))
        for s in cs:
            print(s)

    def test_cs_download(self):
        cs = self.osmo.download_changeset(177967)
        print(cs)

    def test_cs_upload(self):
        diff = '<osmChange version="0.6" generator="pyosm"> ' \
               '<create> ' \
               '<node id="-1" changeset="187558" lat="-33.9135123" lon="151.1173123" /> ' \
               '<node id="-2" changeset="187558" lat="-33.9233721" lon="151.1173321" /> ' \
               '<way id="-3" changeset="187558"> ' \
               '<nd ref="-1"/> <nd ref="-2"/> ' \
               '</way> ' \
               '</create> ' \
               '</osmChange>'
        cs = self.osmo.diff_upload(187558, diff, self.auth())
        print(cs)

    def test_cs_comment(self):
        cs = self.osmo.comm_changeset(177967, 'Hallo Welt', self.auth())
        print(cs)

    def test_get_node(self):
        node = self.osmo.get_element('node', 4314858041)
        print(node)

    def test_get_way(self):
        node = self.osmo.get_element('way', 201774)
        print(node.__repr__())

    def test_get_relation(self):
        rel = self.osmo.get_element('relation', 4304875773)
        print(rel.__repr__())

    def test_get_hist(self):
        node = self.osmo.history_version_element('node', 4314858041)
        print(node.__repr__())

    def test_get_full_elem(self):
        rel = self.osmo.get_full_element('way', 4305504687)
        for item in rel:
            print(item.__repr__)

    def test_get_elem_bbox_pos(self):
        # westlimit=9.3852744541; southlimit=49.1700528219; eastlimit=9.38678722; northlimit=49.1708595043
        print(pyosmapi.osm_util.create_bbox(52.5134, 13.4374, 1000))
        elems = self.osmo.get_element_bbox((13.428416654163087, 52.49863874116848, 13.446383345836914, 52.52816125883152))
        for item in elems:
            print(item.__repr__)

    def test_get_elem_bbox_neg(self):
        with self.assertRaises(exceptions.NoneFoundError):
            # westlimit=9.3852744541; southlimit=49.1700528219; eastlimit=9.38678722; northlimit=49.1708595043
            print(pyosmapi.osm_util.create_bbox(52.5134, 13.4374, 1000))
            elems = self.osmo.get_element_bbox((9.3849565809, 49.1700030402, 9.3867590254, 49.1707360702))
            for item in elems:
                print(item.__repr__)

    def test_get_gpx_bbox(self):
        # 46.7723/12.1855
        print(pyosmapi.osm_util.create_bbox(51.4564, -0.214097, 750))
        gpx = self.osmo.get_gpx_bbox(pyosmapi.osm_util.create_bbox(51.4564, -0.214097, 750), 0)
        print(gpx)

    def test_send_gpx(self):
        gpx = open('/home/marvin/Downloads/2020-05-31_15-23_Sun.gpx').read()
        tid = self.osmo.upload_gpx(gpx, 'test_trace.xml', 'test', {'test', 'osmate'}, self.auth())
        print(tid)

    def test_get_gpx(self):
        tid = self.osmo.get_gpx(1714, self.auth())
        print(tid)

    def test_meta_gpx(self):
        meta = self.osmo.get_meta_gpx(47, self.auth())
        print(meta)

#   only works with defined value
#    def test_delete_gpx(self):
#        self.osmo.delete_gpx()

    def test_get_own_gpx(self):
        list_gpx = self.osmo.get_own_gpx(self.auth())
        print(list_gpx)

    def test_get_notes_bbox(self):
        notes = self.osmo.get_notes_bbox((13.428416654163087, 52.49863874116848, 13.446383345836914, 52.52816125883152))
        print(notes[0])

    def test_get_notes_search(self):
        notes = self.osmo.search_note('abc')
        print(notes[0])

    def test_get_preferences(self):
        prefs = self.osmo.get_own_preferences(self.auth())
        print(prefs)

    def test_put_preferences(self):
        prefs = {'gps.trace.visibility': 'trackable', 'hallo': 'welt'}
        answ = self.osmo.update_own_preferences(prefs, self.auth())
        print(answ)

    def test_get_preference(self):
        prefs = self.osmo.get_own_preference('hello', self.auth())
        print(prefs)

    def test_set_preference(self):
        prefs = self.osmo.set_own_preference('hello', 'welt', self.auth())
        print(prefs)

    def test_delete_preference(self):
        prefs = self.osmo.delete_own_preference('hello', self.auth())
        print(prefs)

    def tests_get_note(self):
        note = self.osmo.get_note(22599)
        print(note)

    def test_close_note(self):
        data = self.osmo.close_note(22599, 'abc', self.auth())
        print(data)

    def test_edit_elem(self):
        node = self.osmo.get_element('node', 4314858041)
        data = self.osmo.edit_element(node, 178488, self.auth())
        print(data)

    def test_get_users(self):
        users = self.osmo.get_users([7634, 7122])
        print(users)


if __name__ == '__main__':
    unittest.main()



# Errors and Exceptions Osmate


class ParseError(Exception):
    def __init__(self, message):
        self.message = message


class ConflictError(Exception):
    def __init__(self, message):
        self.message = message


class MethodError(LookupError):
    def __init__(self, message):
        self.message = message


class NoneFoundError(ValueError):
    def __init__(self, message):
        self.message = message



import requests
from http import HTTPStatus
import logging
import xml.etree.ElementTree as ElemTree
from pyosmapi.osm_util import *
from pyosmapi.exceptions import *

logger = logging.getLogger(__name__)


class OsmApi:
    base_url = 'https://master.apis.dev.openstreetmap.org/api/0.6'

    def __init__(self, live: bool = False):
        if live:
            self.base_url = 'https://api.openstreetmap.org/api/0.6'

    def get_api_versions(self):
        """
        :returns: supported API versions
        """
        data = requests.get('https://api.openstreetmap.org/api' + '/versions')
        if data.ok:
            return ElemTree.fromstring(data.text).find('api/version').text
        raise Exception(data.text)

    def get_api_capabilities(self) -> dict:
        """
        Max / Min values
        image blacklisting
        :returns:
        """
        # allowed by convenience with version
        data = requests.get(self.base_url + '/capabilities')
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            capabiliy = {}
            for item in tree.find('api').iter():
                capabiliy[item.tag] = item.attrib
            blacklist = []
            for item in tree.findall('policy/imagery/blacklist'):
                blacklist.append(item.get('regex'))
            capabiliy['image_blacklist'] = blacklist
            return capabiliy
        raise Exception(data.text)

    def get_permissions(self, auth) -> set:
        """
        current permissions
        Authorisation required else None
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: set of all current permissions
        """
        data = requests.get(self.base_url + '/permissions', auth=auth)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            permissions = set()
            for item in tree.findall('permissions/permission'):
                permissions.add(item.get('name'))
            return permissions if len(permissions) else None
        raise Exception(data.text)

    ############################################### CHANGESET #######################################################

    def open_changeset(self, tags: dict, auth) -> int:
        """
        opens a new changeset and returns its id
        Authorisation required

        :param tags: Dictionary containing additional tags
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: changeset ID
        """
        root = ElemTree.Element('pyosmapi')
        cs = ElemTree.SubElement(root, 'changeset')
        try:
            self.__kv_serial(tags, cs)
        except AttributeError:
            pass
        xml = ElemTree.tostring(root)

        logger.debug(xml)
        data = requests.put(self.base_url + '/changeset/create', data=xml, auth=auth)
        if data.ok:
            return int(data.text)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ParseError(data.text)
        raise Exception(data.text)

    def get_changeset(self, cid: int, discussion: bool = False) -> ChangeSet:
        """
        A Call to get a changeset optionally with discussion.
        no elements included

        :param cid: changeset ID
        :param discussion: include changeset discussion?
        :returns: dictionary representation of the changeset
        :raises NoneFoundError: no changeset matching this ID
        """
        url = self.base_url + '/changeset/{}'.format(cid)
        if discussion:
            url += '?include_discussion=True'
        data = requests.get(url)
        if data.ok:
            logger.debug(data.text)
            tree = ElemTree.fromstring(data.text).find('changeset')
            return self.__changeset_parser(tree)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    def edit_changeset(self, changeset: ChangeSet, auth):
        """
        edits only changeset tags, only tags in this update remain in changeset

        :param changeset: ChangeSet object
        :param auth: either OAuth1 object or tuple (username, password)
        :raises NoneFoundError: no changeset of that ID
        :raises ConflictError: other user than creator trying to use changeset / or changeset already closed.
        """
        root = ElemTree.Element('pyosmapi')
        cs = ElemTree.SubElement(root, 'changeset')
        self.__kv_serial(changeset.tags, cs)
        xml = ElemTree.tostring(root)

        data = requests.put(self.base_url + '/changeset/{}'.format(changeset.id), data=xml, auth=auth)
        if data.ok:
            return None
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def close_changeset(self, cid: int, auth):
        """
        closes a changeset
        Authorisation required

        :param cid: changeset ID
        :param auth: either OAuth1 object or tuple (username, password)
        :raises NoneFoundError: no changeset of that ID
        :raises ConflictError: other user than creator trying to use changeset / or changeset already closed.
        """
        data = requests.get(self.base_url + '/changeset/{}/close'.format(cid), auth=auth)
        if data.ok:
            return None
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def download_changeset(self, cid: int) -> str:
        """
        downloads a OsmChange document

        :param cid: changeset ID
        :raises NoneFoundError: no changeset of that ID
        """
        data = requests.get(self.base_url + '/changeset/{}/download'.format(cid))
        if data.ok:
            return data.text
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    def get_changesets(self, bbox: tuple = None, user: str = '', time: datetime = None,
                       is_open: bool = False, is_closed: bool = False, changesets: list = None) -> list:
        """
        max 100 changesets matching all provided parameters

        :param bbox:(min_lon, min_lat, max_lon, max_lat)
        :param user: username or user_id
        :param time: Time format: Anything that this_ Ruby function will parse.
        :param is_open: xor is_closed
        :param is_closed: xor is_open
        :param changesets: changeset_ids as list
        :returns: found changesets in a list

        .. _this: https://ruby-doc.org/stdlib-2.7.2/libdoc/date/rdoc/DateTime.html#method-c-parse
        """
        params = {}
        if bbox:
            params['bbox'] = ','.join(map(str, bbox))
        if user:
            if user.isdigit():
                params['user'] = int(user)
            else:
                params['display_name'] = user
        if time:
            params['time'] = ','.join(map(str, time.isoformat()))
        if is_open ^ is_closed:
            if is_open:
                params['open'] = True
            else:
                params['closed'] = True
        if changesets:
            params['changesets'] = ','.join(map(str, changesets))

        data = requests.get(self.base_url + '/changesets', params=params)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            cs = []
            for sub in tree.findall('changeset'):
                cs.append(self.__changeset_parser(sub))
            return cs
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError(data.text)
        raise Exception(data.text)

    def diff_upload(self, cid: int, xml: str, auth) -> list:
        """
        uploads all changes at once, an error rolls back all (All or Nothing)

        - each element must carry a changeset and a version attribute, except when you are creating an element where the
          version is not required as the server sets that for you. The changeset must be the same as the
          changeset ID being uploaded to.
        - a <delete> block in the OsmChange document may have an if-unused attribute (the value of which is ignored).
          If this attribute is present, then the delete operation(s) in this block are conditional and will only be
          executed if the object to be deleted is not used by another object. Without the if-unused, such a situation
          would lead to an error, and the whole diff upload would fail. Setting the attribute will also cause
          deletions of already deleted objects to not generate an error.
        - OsmChange documents generally have user and uid attributes on each element.
          These are not required in the document uploaded to the API.

        Authorisation required

        :param cid: changeset id
        :type cid: int/str
        :param xml: OsmChange document as a String
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: list with dict {type, old_id, new_id, new_version}
        """
        data = requests.post(self.base_url + '/changeset/{}/upload'.format(cid), data=xml, auth=auth)
        if data.ok:
            changes = []
            tree = ElemTree.fromstring(data.text)
            for item in tree.iter():
                item: ElemTree.Element
                line = {'type': item.tag, 'old_id': item.get('old_id')}
                if item.get('new_id'):
                    line['new_id'] = item.get('new_id')
                    line['new_version'] = item.get('new_version')
                changes.append(line)
            return changes
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ParseError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def __changeset_parser(self, tree: ElemTree):
        tags = self.__kv_parser(tree.findall('tag'))
        cs_prop = {}
        for key in tree.keys():
            cs_prop[key] = tree.get(key)

        com_xml = tree.findall('discussion/comment')
        comments = []
        for com in com_xml:
            comments.append(Comment(com.find('text').text, com.get('uid'), com.get('user'), com.get('date')))
        try:
            bbox = cs_prop['min_lon'], cs_prop['min_lat'], cs_prop['max_lon'], cs_prop['max_lat']
        except KeyError:
            bbox = (None, None, None, None)
        created = datetime.strptime(cs_prop['created_at'], "%Y-%m-%dT%H:%M:%SZ")
        if cs_prop['created_at'] is True:
            closed = None
        else:
            closed = datetime.strptime(cs_prop['closed_at'], "%Y-%m-%dT%H:%M:%SZ")
        ch_set = ChangeSet(cs_prop['id'], cs_prop['user'], cs_prop['uid'], created,
                           cs_prop['open'], bbox, closed, tags, comments)
        return ch_set

    def comm_changeset(self, cid: int, text: str, auth) -> ChangeSet:
        """
        Add a comment to a changeset. The changeset must be closed.
        Authorisation required

        :param cid: changeset ID
        :param text: text in new comment
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: ChangeSet just commented, no comments
        :raises ValueError: no textfield present
        :raises ConflictError: deleted
        """
        data = requests.post(self.base_url + '/changeset/{}/comment'.format(str(cid)),
                             data={'text': text}, auth=auth)
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text).find('changeset')
            return self.__changeset_parser(tree)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)

    def sub_changeset(self, cid: int, auth) -> ChangeSet:
        """
        Subscribes the current authenticated user to changeset discussion
        Authorisation required

        :param cid: changeset ID
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: ChangeSet just subscribed
        :raises ConflictError: already subscribed
        """
        data = requests.post(self.base_url + '/changeset/{}/subscribe'.format(cid), auth=auth)
        if data.ok:
            tree = ElemTree.fromstring(data.text).find('changeset')
            return self.__changeset_parser(tree)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def unsub_changeset(self, cid: int, auth) -> ChangeSet:
        """
        Unsubscribe the current authenticated user from changeset discussion
        Authorisation required

        :param cid: changeset ID
        :param auth: either OAuth1 object or tuple (username, password)
        :raises NoneFoundError: is not subscribed
        """
        data = requests.post(self.base_url + '/changeset/{}/unsubscribe'.format(cid), auth=auth)
        if data.ok:
            tree = ElemTree.fromstring(data.text).find('changeset')
            return self.__changeset_parser(tree)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    ############################################## ELEMENT #########################################################

    def create_element(self, elem: Element, cid: int, auth) -> int:
        """
        creates new element of specified type
        Authorisation required

        :param elem: element to get created
        :param cid: open changeset ID
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: Element ID
        :raises NoneFoundError:
            - When a changeset ID is missing (non consistent error message)
            - When a node is outside the world
            - When there are too many nodes for a way -> limit, see capabilities
        :raises ConflictError:
            - When changeset already closed
            - When changeset creator and element creator different
        :raises ParseError: When a way/relation has nodes that do not exist or are not visible
        """
        elem.changeset = cid
        xml = self.__serial_elem(elem, True)
        data = requests.get(self.base_url + '/{}/create'.format(elem.e_type), data=xml, auth=auth)
        if data.ok:
            return int(data.text)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        elif data.status_code == HTTPStatus.PRECONDITION_FAILED:
            raise ParseError(data.text)

    def get_element(self, e_type: str, eid: int) -> Element:
        """
        returns only this element

        :param e_type: type of element ('node'/'way'/'relation')
        :param eid: element id
        :returns: requested Element of that Type
        :raises NoneFoundError: No Element with such id
        :raises LockupError: Deleted Element
        """
        data = requests.get(self.base_url + '/{}/{}'.format(e_type, eid))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            return self.__parse_elem(tree[0])
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.GONE:
            raise LookupError(data.text)
        raise Exception(data.text)

    def __parse_elem(self, elem: ElemTree.Element):
        eid = elem.get('id')
        version = int(elem.get('version'))
        changeset = elem.get('changeset')
        cr_date = datetime.strptime(elem.get('timestamp'), "%Y-%m-%dT%H:%M:%SZ")
        user = elem.get('user')
        uid = elem.get('uid')
        visible = bool(elem.get('visible'))
        lat = elem.get('lat')  # lan & lon only type node
        lon = elem.get('lon')

        nodes = []
        for node in elem.findall('nd'):
            nodes.append(node.get('ref'))

        members = []
        for member in elem.findall('member'):
            mem = {}
            for item in member.keys():
                mem[item] = member.get(item)
            members.append(mem)

        if elem.tag == 'node':
            return Node(eid, lat, lon, version, changeset, user, uid, cr_date, visible,
                        self.__kv_parser(elem.findall('tag')))
        elif elem.tag == 'way':
            return Way(eid, nodes, version, changeset, user, uid, cr_date, visible,
                       self.__kv_parser(elem.findall('tag')))
        elif elem.tag == 'relation':
            return Relation(eid, members, version, changeset, user, uid, cr_date, visible,
                            self.__kv_parser(elem.findall('tag')))
        return elem

    def __serial_elem(self, elem: Element, is_create: bool = False) -> str:
        root = ElemTree.Element("pyosmapi")
        doc = ElemTree.Element('None')
        if not is_create:
            params = {'id': elem.id, 'version': str(elem.version), 'changeset': str(elem.changeset),
                      'user': elem.user, 'uid': str(elem.uid),
                      'visible': str(elem.visible), 'timestamp': elem.created}
        else:
            params = {'changeset': str(elem.changeset)}
        if isinstance(elem, Node):
            params['lat'] = str(elem.lat)
            params['lon'] = str(elem.lon)
            doc = ElemTree.SubElement(root, "node", params)
        elif isinstance(elem, Way):
            doc = ElemTree.SubElement(root, "way", params)
            for ref in elem.nodes:
                ElemTree.SubElement(doc, 'nd', {'ref': ref})
        elif isinstance(elem, Relation):
            doc = ElemTree.SubElement(root, "relation", params)
            for member in elem.members:
                ElemTree.SubElement(doc, 'member', member)
        self.__kv_serial(elem.tags, doc)

        return ElemTree.tostring(root).decode()

    def edit_element(self, elem: Element, cid: int, auth) -> int:
        """
        chage tags/position of element
        Authorisation required

        :param elem: changed element to get uploaded (Node/Way/Relation)
        :param cid: open changeset id
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: New version number
        :raises ValueError:
            When a changeset ID is missing
            When a node is outside the world
            When there are too many nodes for a way -> limit, see capabilities
            When the version of the provided element does not match the current database version of the element
        :raises NoneFoundError: Element ID not found
        :raises ConflictError:
            When changeset already closed
            When changeset creator and element creator different
        :raises ParseError: When a way/relation has nodes that do not exist or are not visible
        """
        elem.changeset = cid
        data = requests.put(self.base_url + '/{}/{}'.format(elem.e_type, elem.id),
                            data=self.__serial_elem(elem), auth=auth)
        if data.ok:
            return int(data.text)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        elif data.status_code == HTTPStatus.PRECONDITION_FAILED:
            raise ParseError(data.text)
        raise Exception(data.text)

    def delete_element(self, elem: Element, cid: int, auth) -> int:
        """
        deletes element
        Authorisation required

        :param elem: changed element to get deleted (Node/Way/Relation)
        :param cid: open changeset id
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: new version number
        :raises ValueError:
            When a changeset ID is missing
            When a node is outside the world
            When there are too many nodes for a way
            When the version of the provided element does not match the current database version of the element
        :raises NonFoundError: Element ID not found
        :raises ConflictError:
            When changeset already closed
            When changeset creator and element creator different
        :raises LookupError: deleted
        :raises ParseError:
            When node is still part of way/relation
            When way is still part of relation
            When relation is still part of another relation
        """
        elem.changeset = cid
        data = requests.delete(self.base_url + '/{}/{}'.format(elem.e_type, elem.id),
                               data=self.__serial_elem(elem), auth=auth)
        if data.ok:
            return int(data.text)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        elif data.status_code == HTTPStatus.GONE:
            raise LookupError(data.text)
        elif data.status_code == HTTPStatus.PRECONDITION_FAILED:
            raise ParseError(data.text)
        raise Exception(data.text)

    def history_element(self, e_type: str, eid: int) -> list:
        """
        complete history of an element

        :param e_type: type of element ('node'/'way'/'relation')
        :param eid: element id
        :returns: all versions of that element
        """
        data = requests.get(self.base_url + '/{}/{}/history'.format(e_type, eid))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree:
                elems.append(self.__parse_elem(sub))
            return elems
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    def history_version_element(self, e_type: str, eid: int, version: int = 1) -> Element:
        """
        complete history of an element

        :param e_type: type of element ('node'/'way'/'relation')
        :param eid: element id
        :param version: defaults to 1
        :returns: all versions of that element
        """
        data = requests.get(self.base_url + '/{}/{}/{}'.format(e_type, eid, version))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            return self.__parse_elem(tree[0])
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    def get_elements(self, e_type: str, lst_eid: list) -> list:
        """
        returns multiple elements of same e_type

        :returns: multiple elements as specified in the list of eid
        :param e_type: element type one of [node, way, relation]
        :param lst_eid: list of eid
        :raises ParseError: missing or wrong parameter
        :raises NoneFoundError: requested object never existed
        :raises MethodError: you might never try ro request more than ~700 elements at once
        """
        data = requests.get(self.base_url + '/{}s?{}s={}'.format(e_type, e_type, ','.join(map(str, lst_eid))))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree:
                elems.append(self.__parse_elem(sub))
            return elems

        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ParseError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.REQUEST_URI_TOO_LONG:
            raise MethodError(data.text)
        raise Exception(data.text)

    def get_relation_of_element(self, e_type: str, eid: int) -> list:
        """
        returns all relations containing this element

        :param e_type: type of element ('node'/'way'/'relation')
        :param eid: element id
        :returns: relations containing this element
        :raises NoneFoundError: no such element or no relations containing this element
        """
        data = requests.get(self.base_url + '/{}/{}/relations'.format(e_type, eid))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree:
                elems.append(self.__parse_elem(sub))
            if not elems:
                raise NoneFoundError('no such element or no relations on this element')
            return elems
        raise Exception(data.text)

    def get_ways_of_node(self, eid: int) -> list:
        """
        use only on node elements

        :param eid: element ID
        :returns: ways directly using this node
        :raises NoneFoundError: no connected ways found
        """
        data = requests.get(self.base_url + '/node/{}/ways'.format(eid))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree:
                elems.append(self.__parse_elem(sub))
            if not elems:
                raise NoneFoundError('no such node or no ways on this element')
            return elems
        raise Exception(data.text)

    def get_element_bbox(self, bbox: tuple) -> list:
        """
        all elements within bbox

        :param bbox: tuple (lon_min, lat_min, lon_max, lat_max)
        :returns: all Elements with minimum one Node within this BoundingBox
        :raise NoneFoundError: either none or over 50.000 elements are found
        """
        data = requests.get(self.base_url + '/map?bbox={}'.format(','.join(map(str, bbox))))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree[1:]:
                elems.append(self.__parse_elem(sub))
            if not len(tree) > 1:
                raise NoneFoundError('no elements or over 50.000 elements')
            return elems
        raise Exception(data.text)

    def get_full_element(self, e_type: str, eid: int) -> list:
        """
        returns all elements directly referenced or referenced by 2nd grade

        :param e_type: type of element ('node'/'way'/'relation') referenced
        :param eid: element id
        :returns: elements referenced up to 2nd grade with element
        :raises NoneFoundError: eid not found / element deleted
        """
        data = requests.get(self.base_url + '/{}/{}/full'.format(e_type, eid))
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            logger.debug(data.text)
            elems = []
            for sub in tree:
                elems.append(self.__parse_elem(sub))
            return elems
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.GONE:
            raise NoneFoundError(data.text)
        raise Exception(data.text)

    ################################################## GPX ########################################################

    def get_gpx_bbox(self, bbox: tuple, page: int = 0) -> str:
        """
        returns max 5000GPS trackpoints as gpx string, increase page for any additional 5000

        :param bbox: (min_lon, min_lat, max_lon, max_lat)
        :param page: 5000 trackpoints are returned each page
        :returns: format GPX Version 1.0 string
        """
        data = requests.get(self.base_url + '/trackpoints', params={'bbox': ','.join(map(str, bbox)), 'page': page})
        if data.ok:
            return data.text
        raise Exception(data.text)

    def upload_gpx(self, trace: str, name: str, description: str, tags: set, auth,
                   visibility: str = 'trackable') -> int:
        """
        uploads gpx trace
        Authorisation required

        :param trace: gpx trace file string
        :param description: gpx description
        :param name: file name on pyosmapi
        :param tags: additional tags: eg.: mappingtour, etc
        :param auth: either OAuth1 object or tuple (username, password)
        :param visibility: one of [private, public, trackable, identifiable]
            more https://wiki.openstreetmap.org/wiki/Visibility_of_GPS_traces
        :returns: gpx_id
        """
        content = {'description': description, 'tags': ','.join(tags), 'visibility': visibility}
        req_file = {'file': (name, trace)}
        data = requests.post(self.base_url + '/gpx/create', auth=auth, files=req_file, data=content)
        if data.ok:
            return int(data.text)
        raise Exception(data.text)

    def update_gpx(self, tid: int, trace: str, description: str, tags: list, auth,
                   public: bool = True, visibility: str = 'trackable'):
        """
        updates gpx trace
        Authorisation required

        :param tid: uploaded trace id
        :param trace: gpx trace as string
        :param description: gpx description
        :param tags: additional tags mapping_tour, etc
        :param auth: either OAuth1 object or tuple (username, password)
        :param public: True for public tracks else False
        :param visibility: one of [private, public, trackable, identifiable]
            more https://wiki.openstreetmap.org/wiki/Visibility_of_GPS_traces
        """
        content = {'description': description, 'tags': ','.join(tags), 'public': public, 'visibility': visibility}
        req_file = {'file': ('test-trace.gpx', trace)}
        data = requests.put(self.base_url + '/gpx/' + str(tid), auth=auth, files=req_file, data=content)
        if data.ok:
            logger.debug('updated')
        else:
            logger.debug('not updated')
            raise Exception(data.text)

    def delete_gpx(self, tid: int, auth):
        """
        deletes own gpx indicated by ID
        Authorisation required

        :param tid: trace ID
        :param auth: either OAuth1 object or tuple (username, password)
        """
        data = requests.delete(self.base_url + '/gpx/' + str(tid), auth=auth)
        if data.ok:
            logger.debug('deleted')
            return None
        else:
            logger.debug('not deleted')
            raise Exception(data.text)

    def get_meta_gpx(self, tid: int, auth) -> dict:
        """
        returns meta data of identified gpx
        Authentication required

        :param tid: trace ID
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: dict with metadata
        """
        data = requests.get(self.base_url + '/gpx/{}/details'.format(tid), auth=auth)
        if data.ok:
            logger.debug(data.text)
            tree = ElemTree.fromstring(data.text)
            ret = tree.find('gpx_file').attrib
            ret['timestamp'] = datetime.strptime(ret['timestamp'], "%Y-%m-%dT%H:%M:%SZ")
            ret['description'] = tree.find('gpx_file/description').text
            tags = []
            for tag in tree.findall('gpx_file/tag'):
                tags.append(tag.text)
            ret['tags'] = tags
            return ret
        raise Exception(data.text)

    def get_gpx(self, tid: int, auth) -> str:
        """identifying the gpx file
        downloads public or own private gpx file
        Authentication required

        :param tid: trace ID
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: the full gpx file as a string
        """
        data = requests.get(self.base_url + '/gpx/{}/data'.format(tid), auth=auth)
        if data.ok:
            return data.text
        raise Exception(data.text)

    def get_own_gpx(self, auth) -> list:
        """
        meta data of all own gpx files
        Authorisation required

        :param auth: either OAuth1 object or tuple (username, password)
        :returns: list of dictionary representing the metadata
        """
        data = requests.get(self.base_url + '/user/gpx_files', auth=auth)
        if data.ok:
            return self.__parse_gpx_info(data.text)
        raise Exception(data.text)

    def __parse_gpx_info(self, xml):
        tree = ElemTree.fromstring(xml)
        lst = []
        for item in tree.findall('gpx_file'):
            attrib = item.attrib
            attrib['timestamp'] = datetime.strptime(attrib['timestamp'], "%Y-%m-%dT%H:%M:%SZ")
            for info in item:
                attrib[info.tag] = info.text
            lst.append(attrib)
        return lst

    ################################################# USER ######################################################

    def get_user(self, uid: int) -> dict:
        """
        user data and  statistic

        :param uid: user ID
        :returns: dictionary with user detail
        """
        data = requests.get(self.base_url + '/user/' + str(uid))
        if data.ok:
            return self.__parse_user(data.text)[0]
        raise Exception(data.text)

    def get_users(self, uids: list) -> list:
        """
        user data an statistics for multiple users

        :param uids: uid in a list
        :returns: list of dictionary with user detail
        """
        data = requests.get(self.base_url + '/users?users=' + ','.join(map(str, uids)))
        if data.ok:
            logger.debug(data.text)
            return self.__parse_user(data.text)
        raise Exception(data.text)

    def get_current_user(self, auth) -> dict:
        """
        own user data and statistics
        Authorisation required

        :param auth: either OAuth1 object or tuple (username, password)
        :returns: dictionary with user detail
        """
        data = requests.get(self.base_url + '/user/details', auth=auth)
        if data.ok:
            return self.__parse_user(data.text)[0]
        raise Exception(data.text)

    def __parse_user(self, xml: str) -> list:
        tree = ElemTree.fromstring(xml)
        users = []
        for user in tree.findall('user'):
            users.append({'uid': user.get('id'),
                          'name': user.get('display_name'),
                          'cr_date': user.get('account_created'),
                          'description': user.find('description').text,
                          'terms': bool(user.find('contributor-terms').get('agreed')),
                          'changeset_count': int(user.find('changesets').get('count')),
                          'traces_count': int(user.find('traces').get('count'))})
        return users

    def get_own_preferences(self, auth) -> dict:
        """
        returns all own user preferences
        Authorisation required

        :param auth: either OAuth1 object or tuple (username, password)
        :returns: dictionary with preferences
        """
        data = requests.get(self.base_url + '/user/preferences', auth=auth)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__kv_parser(tree.findall('preferences/preference'))
        raise Exception(data.text)

    def update_own_preferences(self, pref: dict, auth):
        """
        updates all user preferences at once
        Authorisation required

        :param auth: either OAuth1 object or tuple (username, password)
        :param pref: dictionary with preferences
        """
        root = ElemTree.Element("pyosmapi")
        prefs = ElemTree.SubElement(root, 'preferences')
        for key, value in pref.items():
            ElemTree.SubElement(prefs, 'preference', {'k': key, 'v': value})
        data = ElemTree.tostring(root).decode()
        print(data)
        ret = requests.put(self.base_url + '/user/preferences', data=data, auth=auth)
        if ret.ok:
            return None
        raise Exception(ret.text)

    def get_own_preference(self, key: str, auth) -> str:
        """
        returns the value of a single preference
        Authorisation required

        GET /api/0.6/user/preferences/<your_key>

        :param auth: either OAuth1 object or tuple (username, password)
        :param key: key of preference
        :returns: value
        """
        data = requests.get(self.base_url + '/user/preferences/{}'.format(key), auth=auth)
        if data.ok:
            return data.text
        raise Exception(data.text)

    def set_own_preference(self, key: str, value: str, auth):
        """
        updates the value of a single preference
        Authorisation required

        :param key: key of preference
        :param value: new value
        :param auth: either OAuth1 object or tuple (username, password)
        """
        data = requests.put(self.base_url + '/user/preferences/{}'.format(key), data=value, auth=auth)
        if data.ok:
            return None
        raise Exception(data.text)

    def delete_own_preference(self, key: str, auth):
        """
        deletes a single preference
        Authorisation required

        :param key: key of preference
        :param auth: either OAuth1 object or tuple (username, password)
        """
        data = requests.delete(self.base_url + '/user/preferences/{}'.format(key), auth=auth)
        if data.ok:
            return None
        raise Exception(data.text)

    ################################################# NOTE ######################################################

    def __parse_notes(self, tree: ElemTree.Element):
        lst = []
        for item in tree.findall('note'):
            lon = item.get('lon')
            lat = item.get('lat')
            nid = item.find('id').text
            main_created = datetime.strptime(item.find('date_created').text, "%Y-%m-%d %H:%M:%S UTC")
            is_open = item.find('status').text
            if not is_open == 'closed':
                is_open = True
            else:
                is_open = False
                datetime.strptime(item.find('date_closed').text, "%Y-%m-%d %H:%M:%S UTC")

            comments = []
            for comment in item.findall('comments/comment'):
                created = datetime.strptime(comment.find('date').text, "%Y-%m-%d %H:%M:%S UTC")
                uid = comment.find('uid').text
                user = comment.find('user').text
                text = comment.find('text').text
                action = comment.find('action').text
                comments.append(Comment(text, uid, user, created, action))
            lst.append(Note(nid, lat, lon, main_created, is_open, comments))
        return lst

    def get_notes_bbox(self, bbox: tuple, limit: int = 100, closed: int = 7) -> list:
        """
        searches for all notes within the boundaries of bbox

        :param bbox: (lon_min, lat_min, lon_max, lat_max)
        :param limit: 0-1000
        :param closed: max days closed -1=all, 0=only_open
        :returns: list of Notes
        :raises ValueError: When any of the limits are crossed
        """
        data = requests.get(self.base_url + '/notes', params={'bbox': ','.join(map(str, bbox)),
                                                              'limit': limit, 'closed': closed})
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        raise Exception(data.text)

    def get_note(self, nid: int) -> Note:
        """
        a note with all comments

        :param nid: note id
        :return: the identified Note
        :raises NoneFoundError: note ID not found
        """
        data = requests.get(self.base_url + '/notes/{}'.format(str(nid)))
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)[0]
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        raise Exception(data.text)

    def create_note(self, text: str, lat: float, lon: float, auth) -> Note:
        """
        creates anonymous or user made notes depending if auth is provided
        Authorisation optional

        :param text: Note Text
        :param lat: latitude
        :param lon: longitude
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: note ID
        :raises ValueError: No text field
        """
        data = requests.post(self.base_url + '/notes', params={'lat': lat, 'lon': lon, 'text': text}, auth=auth)
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)[0]
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        raise Exception(data.text)

    def comment_note(self, nid: int, text: str, auth) -> Note:
        """
        adds a comment to the note
        Authorisation required

        :param nid: note ID
        :param text: Text
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: the Note itself
        :raises ValueError: no Textfield
        :raises NoneFoundError: note ID not found
        :raises ConflictError: already closed Note
        """
        data = requests.post(self.base_url + '/notes/{}/comment'.format(str(nid)),
                             params={'text': text}, auth=auth)
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)[0]
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def close_note(self, nid: int, text: str, auth) -> Note:
        """
        closes a note, no comments can be added to a closed note
        Authorisation required

        :param nid: note ID
        :param text: Closing comment
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: the Note itself
        :raises NoneFoundError: note ID not found
        :raises ConflictError: already closed Note
        """
        data = requests.post(self.base_url + '/notes/{}/close'.format(str(nid)),
                             params={'text': text}, auth=auth)
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)[0]
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        raise Exception(data.text)

    def reopen_note(self, nid: int, text: str, auth):
        """
        reopens a note, open for more comments
        Authorisation required

        :param nid: Note ID
        :param text: Text
        :param auth: either OAuth1 object or tuple (username, password)
        :returns: the Note itself
        :raises NoneFoundError: note ID not found
        :raises ConflictError: already closed Note
        :raises LookupError: deleted Note
        """
        data = requests.post(self.base_url + '/notes/{}/close'.format(str(nid)),
                             params={'text': text}, auth=auth)
        logger.debug(data.text)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)[0]
        elif data.status_code == HTTPStatus.NOT_FOUND:
            raise NoneFoundError(data.text)
        elif data.status_code == HTTPStatus.CONFLICT:
            raise ConflictError(data.text)
        elif data.status_code == HTTPStatus.GONE:
            raise LookupError(data.text)
        raise Exception(data.text)

    def search_note(self, text: str, limit: int = 100, closed: int = 7, user: str = None,
                    start: datetime = None, end: datetime = None,
                    sort: str = 'updated_at', order: str = 'newest') -> list:
        """
        searches for notes complying all parameters
        Authorisation required

        :param text: <free text>
        :param limit: 0-1000 max amount notes returned
        :param closed: max days closed -1=all, 0=only_open
        :param user: User ID or Username
        :param start: from earliest date
        :param end: to newer date default: today
        :param sort: created_at or updated_at
        :param order: oldest or newest
        :returns: list of Notes
        :raises ValueError: When any of the limits are crossed
        """
        params = {'q': text, 'limit': limit, 'closed': closed, 'sort': sort, 'order': order}
        if user:
            if type(user) == int or (type(user) == str and user.isdigit()):
                params['user'] = int(user)
            else:
                params['username'] = user
        if start:
            params['from'] = start.isoformat()
        if end:
            params['to'] = end.isoformat()

        data = requests.get(self.base_url + '/notes/search', params=params)
        if data.ok:
            tree = ElemTree.fromstring(data.text)
            return self.__parse_notes(tree)
        elif data.status_code == HTTPStatus.BAD_REQUEST:
            raise ValueError(data.text)

    def rss_notes(self, bbox) -> str:
        """
        gets a notes RSS feed of the area

        :param bbox: (lon_min, lat_min, lon_max, lat_max)
        :return: xml RSS feed
        """
        params = {'bbox': ','.join(map(str, bbox))}
        data = requests.get(self.base_url + '/notes/feed', params=params)
        if data.ok:
            return data.text
        raise Exception(data.text)

    def __kv_parser(self, lst: list) -> dict:
        """
        :param lst: list of tags form <tag k="some" v="value"/>
        :return: dictionary of key value pairs
        """
        tags = {}
        for item in lst:
            tags[item.get('k')] = item.get('v')
        return tags

    def __kv_serial(self, tags: dict, parent: ElemTree.Element):
        for key, value in tags.items():
            ElemTree.SubElement(parent, 'tag', {'k': key, 'v': value})



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyosmapi",
    version="0.0.1",
    author="jonycoo",
    author_email="jonathan.koehler.7@web.de",
    description="Open Street Map API Wrapper in Python",
    keywords='osm, openstreetmap, wrapper, api',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/jonycoo/py-osmapi",
    packages=setuptools.find_packages(),
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
          'requests'
      ],
    python_requires='>=3.6',
)



import json
import math
from datetime import datetime

OSM_URL = 'https://master.apis.dev.openstreetmap.org'


class Element:
    def __init__(self, eid: int, version: int, changeset: int,
                 user: str, uid: int, created: datetime, visible: bool, tags: dict):
        self._id = eid
        self.tags = tags or {}
        self.version = version
        self.changeset = changeset
        self.user = user
        self.uid = uid
        self.created = created
        self.visible = visible
        self.e_type = 'element'

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda o: _default(o))

    def __str__(self):

        str_tags = ''
        for attr, val in self.tags.items():
            str_tags += f"\n\t{attr} -- {val}"

        return f"type:{self.type}\nid:{self.id}\ntags:{str_tags}\nversion: {self.version}\n" \
               f"created: {self.created}\nvisible: {self.visible}" \
               f"\nchangeset:\n\tchangeset_id: {self.changeset}\n\tuser: {self.user}\n\tuser_id: {self.uid}\n"

    @property
    def id(self):
        return self._id

    @property
    def type(self):
        return self.e_type


class Node(Element):
    def __init__(self, eid: int, lat: float, lon: float, version: int, changeset: int,
                 user: str, uid: int, created: datetime, visible: bool, tags: dict):
        super().__init__(eid, version, changeset, user, uid, created, visible, tags)
        self.lat = lat
        self.lon = lon
        self.e_type = 'node'

    def __str__(self):
        return super(Node, self).__str__() + f"location:\n\tlatitude: {self.lat}\n\tlongitude: {self.lon}"


class Way(Element):
    def __init__(self, eid: int, nodes: list, version: int, changeset: int,
                 user: str, uid: int, created: datetime, visible: bool, tags: dict):
        super().__init__(eid, version, changeset, user, uid, created, visible, tags)
        self.nodes = nodes
        self.e_type = 'way'

    def __str__(self):
        return super(Way, self).__str__() + f"nodes: \n\t{self.nodes}"


class Relation(Element):
    def __init__(self, eid: int, members: list, version: int, changeset: int,
                 user: str, uid: int, created: datetime, visible: bool, tags: dict):
        super().__init__(eid, version, changeset, user, uid, created, visible, tags)
        self.members = members
        self.e_type = 'relation'

    def __str__(self):
        return super(Relation, self).__str__() + f"nodes: \n\t{self.members}"


class Comment:
    def __init__(self, text: str, uid: int, username: str, created: datetime, action: str = None):
        self.text = text
        self._id = uid
        self.user = username
        self.created = created
        self.action = action

    def __repr__(self):
        return json.dumps(self.__dict__)

    def __str__(self):
        return f"\n\tid: {self.id}\n\tusername: {self.user}\n\t" \
               f"text: \"{self.text}\"\n\taction: {self.action}\n\tcreated: {self.created}"

    @property
    def id(self):
        return self._id


class Note:
    def __init__(self, nid: int, lat: float, lon: float, created: datetime, is_open: bool,
                 comments: list, closed: datetime = None):
        self._id = nid
        self.lat = lat
        self.lon = lon
        self.open = is_open
        self.comments = comments
        self.created = created
        self.closed = closed

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda o: _default(o))

    def __str__(self):
        com_str = ''
        for com in self.comments:
            com_str += f'{com}'
        return f"id: {self.id}\nlocation:\n\tlatitude: {self.lat}\n\tlongitude: {self.lon}\nopen: {self.open}\n" \
               f"created: {self.created}{f'closed: {self.closed}' if not self.open else ''}\ncomments:{com_str}"

    @property
    def id(self):
        return self._id

    @property
    def location(self):
        return self.lat, self.lon


class Trace:
    def __init__(self, tid: int, gpx: str, filename: str, username: str, created: datetime,
                 desc: str = None, tags: set = None, visibility: str = 'trackable'):
        self._id = tid
        self.gpx = gpx
        self.name = filename
        self.username = username
        self.created = created
        self.desc = desc or 'no Description'
        self.tags = tags or []
        self.visibility = visibility

    def __repr__(self):
        return json.dumps(self.__dict__)

    def __str__(self):
        return f"id: {self.id}\nfilename: {self.name}\nDescription: {self.desc}\ntags: {self.tags}\n" \
               f"date: {self.created}\nvisible: {self.visibility}\nuploader: {self.username}"

    @property
    def id(self):
        return self._id

    @staticmethod
    def url(osm_user, tid):
        return OSM_URL + '/user/' + osm_user + '/traces/' + tid


class ChangeSet:
    def __init__(self, cid: int, username: str, uid: int, created: datetime, is_open: bool, bbox: tuple,
                 closed: datetime = None, tags: dict = None, comments: list = None):
        """
        :param bbox (min_lon, min_lat, max_lon, max_lat)
        """
        self._id = cid
        self.user = username
        self.uid = uid
        self.created = created
        self.open = is_open
        self.min_lon = bbox[0] or None
        self.min_lat = bbox[1] or None
        self.max_lon = bbox[2] or None
        self.max_lat = bbox[3] or None
        self.closed = closed
        self.tags = tags or {}
        self.comments = comments or []

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda o: _default(o))

    def __str__(self):
        com_str = ''
        for com in self.comments:
            com_str += f'{com}'
        tag_str = ''
        for tag, val in self.tags.items():
            tag_str += f'\n\t{tag}: {val}'
        return f"id: {self.id}\ncreator: {self.user}\ncreator_id: {self.uid}\ncreator_date: {self.created}\n" \
               f"open: {self.open}\nbbox: {(self.min_lat, self.min_lon, self.max_lat, self.max_lon)}\n" \
               f"closed: {self.closed}\ntags: {tag_str}\ncomments: {com_str}"

    @property
    def id(self):
        return self._id

    @property
    def open_comment(self):
        try:
            return self.tags['comment']
        except KeyError:
            return None

    @property
    def bbox(self):
        return self.min_lon, self.min_lat, self.max_lon, self.max_lat


EARTH_RAD = float(6378000)


def create_bbox(lat: float, lon: float, rad: int):
    """
    creates a Geo Bounding box with lat, lon as center

    :param lat: latitude
    :param lon: longitude
    :param rad: radius in meters
    """

    lat_d = (math.asin(float(rad) / (EARTH_RAD * math.cos(math.pi * lat / 180)))) * 180 / math.pi
    lon_d = (math.asin(float(rad) / EARTH_RAD)) * 180 / math.pi

    dec = 8

    max_lat = round(lat + lat_d, dec)
    min_lat = round(lat - lat_d, dec)
    max_lon = round(lon + lon_d, dec)
    min_lon = round(lon - lon_d, dec)

    return min_lon, min_lat, max_lon, max_lat


def _default(obj):
    """Default JSON serializer."""
    import calendar

    if isinstance(obj, datetime):
        if obj.utcoffset() is not None:
            obj = obj - obj.utcoffset()
        millis = int(
            calendar.timegm(obj.timetuple()) * 1000 +
            obj.microsecond / 1000
        )
        return millis
    else:
        return obj.__dict__
    # raise TypeError('Not sure how to serialize %s' % (obj,))
